@@ -1,5 +1,6 @@
 use crate::config::{Config, Unsafe, WriteTypes};
 use crate::converter::{CSharpTypeConverter, Converter};
+use crate::overloads::{Helper, OverloadWriter};
 use interoptopus::lang::c::{CType, CompositeType, Constant, Documentation, EnumType, Field, FnPointerType, Function, Meta, PrimitiveType, Variant, Visibility};
 use interoptopus::patterns::api_guard::library_hash;
 use interoptopus::patterns::callbacks::NamedCallback;
@@ -19,6 +20,15 @@ pub trait CSharpWriter {
 
     fn converter(&self) -> &Converter;
 
+    fn overloads(&self) -> &[Box<dyn OverloadWriter>];
+
+    fn helper(&self) -> Helper {
+        Helper {
+            config: self.config(),
+            converter: self.converter(),
+        }
+    }
+
     fn write_file_header_comments(&self, w: &mut IndentWriter) -> Result<(), Error> {
         indented!(w, r#"{}"#, &self.config().file_header_comment)?;
         Ok(())
@@ -34,17 +44,15 @@ pub trait CSharpWriter {
 
     fn write_imports(&self, w: &mut IndentWriter) -> Result<(), Error> {
         self.debug(w, "write_imports")?;
+
+        indented!(w, r#"#pragma warning disable 0105"#)?;
         indented!(w, r#"using System;"#)?;
         indented!(w, r#"using System.Collections;"#)?;
         indented!(w, r#"using System.Collections.Generic;"#)?;
         indented!(w, r#"using System.Runtime.InteropServices;"#)?;
 
-        if self.config().use_unsafe == Unsafe::UnsafePlatformMemCpy {
-            indented!(w, r#"#if UNITY_2018_1_OR_NEWER"#)?;
-            indented!(w, r#"using Unity.Collections.LowLevel.Unsafe;"#)?;
-            indented!(w, r#"#else"#)?;
-            indented!(w, r#"using System.Runtime.CompilerServices;"#)?;
-            indented!(w, r#"#endif"#)?;
+        for overload in self.overloads() {
+            overload.write_imports(w, self.helper())?;
         }
 
         for namespace_id in self.library().namespaces() {
@@ -56,6 +64,7 @@ pub trait CSharpWriter {
 
             indented!(w, r#"using {};"#, namespace)?;
         }
+        indented!(w, r#"#pragma warning restore 0105"#)?;
 
         Ok(())
     }
@@ -129,9 +138,10 @@ pub trait CSharpWriter {
         self.write_function_annotation(w, function)?;
         self.write_function_declaration(w, function)?;
 
-        w.newline()?;
-
-        self.write_function_overloaded(w, function)?;
+        for overload in self.overloads() {
+            w.newline()?;
+            overload.write_function_overload(w, self.helper(), function)?;
+        }
 
         Ok(())
     }
@@ -167,103 +177,6 @@ pub trait CSharpWriter {
         indented!(w, r#"public static extern {} {}({});"#, rval, name, params.join(", "))
     }
 
-    #[rustfmt::skip]
-    fn write_function_overloaded(&self, w: &mut IndentWriter, function: &Function) -> Result<(), Error> {
-        self.debug(w, "write_function_overloaded")?;
-
-        // If there is nothing to write, don't do it
-        if !self.converter().has_overloadable(function.signature()) {
-            return Ok(());
-        }
-
-        let mut to_pin_name = Vec::new();
-        let mut to_pin_slice_type = Vec::new();
-        let mut to_invoke = Vec::new();
-        let rval = self.converter().function_rval_to_csharp_typename(function);
-        let name = self.converter().function_name_to_csharp_name(function);
-
-        let mut params = Vec::new();
-        for (_, p) in function.signature().params().iter().enumerate() {
-            let name = p.name();
-            let native = self.converter().pattern_to_native_in_signature(p, function.signature());
-            let the_type = self.converter().function_parameter_to_csharp_typename(p, function);
-
-            match p.the_type() {
-                CType::Pattern(TypePattern::Slice(_) | TypePattern::SliceMut(_)) => {
-                    to_pin_name.push(name);
-                    to_pin_slice_type.push(the_type);
-                    to_invoke.push(format!("{}_slice", name));
-                }
-                _ => {
-                    if native.contains("out ") {
-                        to_invoke.push(format!("out {}", name.to_string()));
-                    } else if native.contains("ref ") {
-                        to_invoke.push(format!("ref {}", name.to_string()));
-                    } else {
-                        to_invoke.push(name.to_string());
-                    }
-                }
-            }
-
-            params.push(format!("{} {}", native, name));
-        }
-
-
-        let return_stmt = if function.signature().rval().is_void() { "" } else { "return " };
-
-        indented!(w, r#"public static {} {}({}) {{"#, rval, name, params.join(", "))?;
-
-        if self.config().use_unsafe.any_unsafe() {
-
-            // unsafe
-            //     {
-            //         fixed (Vec3f32* ptr = ffi_slice)
-            //         {
-            //             var ffi_slice_slice = new SliceVec3f32(new IntPtr(ptr), (ulong)ffi_slice.Length);
-            //             return pattern_ffi_slice_2(ffi_slice_slice, i);
-            //         }
-            //     }
-
-            indented!(w, [_], r#"unsafe"#)?;
-            indented!(w, [_], r#"{{"#)?;
-            w.indent();
-
-            for (pin_var, slice_struct) in to_pin_name.iter().zip(to_pin_slice_type.iter()) {
-                indented!(w, [_], r#"fixed (void* ptr_{} = {})"#, pin_var, pin_var)?;
-                indented!(w, [_], r#"{{"#)?;
-                indented!(w, [_ _], r#"var {}_slice = new {}(new IntPtr(ptr_{}), (ulong) {}.Length);"#, pin_var, slice_struct, pin_var, pin_var)?;
-                w.indent();
-            }
-
-            indented!(w, [_], r#"{}{}({});"#, return_stmt, name, to_invoke.join(", "))?;
-            for _ in to_pin_name.iter() {
-                w.unindent();
-                indented!(w, [_], r#"}}"#)?;
-            }
-
-            w.unindent();
-            indented!(w, [_], r#"}}"#)?;
-        } else {
-            for (pin_var, slice_struct) in to_pin_name.iter().zip(to_pin_slice_type.iter()) {
-                indented!(w, [_], r#"var {}_pinned = GCHandle.Alloc({}, GCHandleType.Pinned);"#, pin_var, pin_var)?;
-                indented!(w, [_], r#"var {}_slice = new {}({}_pinned, (ulong) {}.Length);"#, pin_var, slice_struct, pin_var, pin_var)?;
-            }
-
-            indented!(w, [_], r#"try"#)?;
-            indented!(w, [_], r#"{{"#)?;
-            indented!(w, [_ _], r#"{}{}({});"#, return_stmt, name, to_invoke.join(", "))?;
-            indented!(w, [_], r#"}}"#)?;
-            indented!(w, [_], r#"finally"#)?;
-            indented!(w, [_], r#"{{"#)?;
-            for pin in &to_pin_name {
-                indented!(w, [_ _], r#"{}_pinned.Free();"#, pin)?;
-            }
-            indented!(w, [_], r#"}}"#)?;
-        }
-
-        indented!(w, r#"}}"#)
-    }
-
     fn write_type_definitions(&self, w: &mut IndentWriter) -> Result<(), Error> {
         for the_type in self.library().ctypes() {
             self.write_type_definition(w, the_type)?;
@@ -653,6 +566,10 @@ pub trait CSharpWriter {
         indented!(w, [_ _], r#"this.len = count;"#)?;
         indented!(w, [_], r#"}}"#)?;
 
+        for overload in self.overloads() {
+            overload.write_pattern_slice_overload(w, self.helper(), context_type_name, &type_string)?;
+        }
+
         // Getter
         indented!(w, [_], r#"public {} this[int i]"#, type_string)?;
         indented!(w, [_], r#"{{"#)?;
@@ -687,10 +604,16 @@ pub trait CSharpWriter {
             indented!(w, [_ _ _], r#"{{"#)?;
             indented!(w, [_ _ _ _ ], r#"fixed (void* dst = rval)"#)?;
             indented!(w, [_ _ _ _ ], r#"{{"#)?;
-            indented!(w, [_ _ _ _ _], r#"#if UNITY_2018_1_OR_NEWER"#)?;
-            indented!(w, [_ _ _ _ _], r#"UnsafeUtility.MemCpy(dst, data.ToPointer(), (long) (len * (ulong) sizeof({})));"#, type_string)?;
+            indented!(w, [_ _ _ _ _], r#"#if __INTEROPTOPUS_NEVER"#)?;
+
+            for overload in self.overloads() {
+                overload.write_pattern_slice_unsafe_copied_fragment(w, self.helper(), &type_string)?;
+            }
+
             indented!(w, [_ _ _ _ _], r#"#else"#)?;
-            indented!(w, [_ _ _ _ _], r#"Unsafe.CopyBlock(dst, data.ToPointer(), (uint)len);"#)?;
+            indented!(w, [_ _ _ _ _], r#"for (var i = 0; i < (int) len; i++) {{"#)?;
+            indented!(w, [_ _ _ _ _ _], r#"rval[i] = this[i];"#)?;
+            indented!(w, [_ _ _ _ _], r#"}}"#)?;
             indented!(w, [_ _ _ _ _], r#"#endif"#)?;
             indented!(w, [_ _ _ _ ], r#"}}"#)?;
             indented!(w, [_ _ _], r#"}}"#)?;
@@ -758,6 +681,10 @@ pub trait CSharpWriter {
         indented!(w, [_ _], r#"this.len = count;"#)?;
         indented!(w, [_], r#"}}"#)?;
 
+        for overload in self.overloads() {
+            overload.write_pattern_slice_overload(w, self.helper(), context_type_name, &type_string)?;
+        }
+
         // Getter
         indented!(w, [_], r#"public {} this[int i]"#, type_string)?;
         indented!(w, [_], r#"{{"#)?;
@@ -805,10 +732,16 @@ pub trait CSharpWriter {
             indented!(w, [_ _ _], r#"{{"#)?;
             indented!(w, [_ _ _ _ ], r#"fixed (void* dst = rval)"#)?;
             indented!(w, [_ _ _ _ ], r#"{{"#)?;
-            indented!(w, [_ _ _ _ _], r#"#if UNITY_2018_1_OR_NEWER"#)?;
-            indented!(w, [_ _ _ _ _], r#"UnsafeUtility.MemCpy(dst, data.ToPointer(), (long) (len * (ulong) sizeof({})));"#, type_string)?;
+            indented!(w, [_ _ _ _ _], r#"#if __FALSE"#)?;
+
+            for overload in self.overloads() {
+                overload.write_pattern_slice_unsafe_copied_fragment(w, self.helper(), &type_string)?;
+            }
+
             indented!(w, [_ _ _ _ _], r#"#else"#)?;
-            indented!(w, [_ _ _ _ _], r#"Unsafe.CopyBlock(dst, data.ToPointer(), (uint)len);"#)?;
+            indented!(w, [_ _ _ _ _], r#"for (var i = 0; i < (int) len; i++) {{"#)?;
+            indented!(w, [_ _ _ _ _ _], r#"rval[i] = this[i];"#)?;
+            indented!(w, [_ _ _ _ _], r#"}}"#)?;
             indented!(w, [_ _ _ _ _], r#"#endif"#)?;
             indented!(w, [_ _ _ _ ], r#"}}"#)?;
             indented!(w, [_ _ _], r#"}}"#)?;
@@ -863,53 +796,35 @@ pub trait CSharpWriter {
 
         for ctor in class.constructors() {
             // Ctor
-            let args = self.pattern_class_args_without_first_to_string(ctor, true);
             self.write_documentation(w, ctor.meta().documentation())?;
-            indented!(w, r#"public {}({})"#, context_type_name, args)?;
-            indented!(w, r#"{{"#)?;
-            w.indent();
-            self.write_pattern_service_success_enum_aware_rval(w, class, ctor, false)?;
-            w.unindent();
-            indented!(w, r#"}}"#)?;
+            self.write_pattern_service_method(w, class, ctor, "", context_type_name, true)?;
             w.newline()?;
         }
 
         // Dtor
-        indented!(w, r#"public void Dispose()"#)?;
-        indented!(w, r#"{{"#)?;
-        w.indent();
-        self.write_pattern_service_success_enum_aware_rval(w, class, class.destructor(), false)?;
-        w.unindent();
-        indented!(w, r#"}}"#)?;
+        self.write_pattern_service_method(w, class, class.destructor(), "void", "Dispose", true)?;
         w.newline()?;
 
         for function in class.methods() {
             // Main function
-            let args = self.pattern_class_args_without_first_to_string(function, true);
             let without_common_prefix = function.name().replace(&common_prefix, "");
             let prettified = IdPrettifier::from_rust_lower(&without_common_prefix);
             let fn_name = prettified.to_camel_case();
+
+            // Write checked method. These are "normal" methods that accept
+            // common C# types.
             let rval = match function.signature().rval() {
                 CType::Pattern(TypePattern::FFIErrorEnum(_)) => "void".to_string(),
                 _ => self.converter().to_typespecifier_in_rval(function.signature().rval()),
             };
-
             self.write_documentation(w, function.meta().documentation())?;
+            self.write_pattern_service_method(w, class, function, &rval, &fn_name, false)?;
 
-            indented!(w, r#"public {} {}({})"#, rval, fn_name, &args)?;
-            indented!(w, r#"{{"#)?;
-            w.indent();
-            self.write_pattern_service_success_enum_aware_rval(w, class, function, true)?;
-            w.unindent();
-            indented!(w, r#"}}"#)?;
-            w.newline()?;
-
-            // Overloaded
-            if self.converter().has_overloadable(function.signature()) {
-                self.write_documentation(w, function.meta().documentation())?;
-                self.write_pattern_service_method_overload(w, class, function, &rval, &fn_name)?;
-                w.newline()?;
+            for overload in self.overloads() {
+                overload.write_service_method_overload(w, self.helper(), class, function, &fn_name)?;
             }
+
+            w.newline()?;
         }
 
         indented!(w, r#"public IntPtr Context => _context;"#)?;
@@ -922,16 +837,32 @@ pub trait CSharpWriter {
         Ok(())
     }
 
-    fn write_pattern_service_method_overload(&self, w: &mut IndentWriter, _class: &Service, function: &Function, rval: &str, fn_name: &str) -> Result<(), Error> {
-        self.debug(w, "write_pattern_service_method_overload")?;
+    #[allow(clippy::too_many_arguments)]
+    fn write_pattern_service_method(
+        &self,
+        w: &mut IndentWriter,
+        _class: &Service,
+        function: &Function,
+        rval: &str,
+        fn_name: &str,
+        write_contxt_by_ref: bool,
+    ) -> Result<(), Error> {
+        self.debug(w, "write_pattern_service_method")?;
+
         let mut names = Vec::new();
         let mut to_invoke = Vec::new();
         let mut types = Vec::new();
 
+        // For every parameter except the first, figure out how we should forward
+        // it to the invocation we perform.
         for p in function.signature().params().iter().skip(1) {
             let name = p.name();
-            let native = self.converter().pattern_to_native_in_signature(p, function.signature());
 
+            // If we call the checked function we want to resolve a `SliceU8` to a `byte[]`,
+            // but if we call the unchecked version we want to keep that `Sliceu8` in our signature.
+            let native = self.converter().to_typespecifier_in_param(p.the_type());
+
+            // Forward `ref` and `out` accordingly.
             if native.contains("out ") {
                 to_invoke.push(format!("out {}", name.to_string()));
             } else if native.contains("ref ") {
@@ -944,12 +875,23 @@ pub trait CSharpWriter {
             types.push(native);
         }
 
-        let arg_tokens = names.iter().zip(types.iter()).map(|(n, t)| format!("{} {}", t, n)).collect::<Vec<_>>();
-        let fn_call = format!(r#"{}.{}(_context, {})"#, self.config().class, function.name(), to_invoke.join(", "));
+        let method_to_invoke = function.name().to_string();
+        let extra_args = if to_invoke.is_empty() {
+            "".to_string()
+        } else {
+            format!(", {}", to_invoke.join(", "))
+        };
 
+        // Assemble actual function call.
+        let context = if write_contxt_by_ref { "ref _context" } else { "_context" };
+        let arg_tokens = names.iter().zip(types.iter()).map(|(n, t)| format!("{} {}", t, n)).collect::<Vec<_>>();
+        let fn_call = format!(r#"{}.{}({}{})"#, self.config().class, method_to_invoke, context, extra_args);
+
+        // Write signature.
         indented!(w, r#"public {} {}({})"#, rval, fn_name, arg_tokens.join(", "))?;
         indented!(w, r#"{{"#)?;
 
+        // Determine return value behavior and write function call.
         match function.signature().rval() {
             CType::Pattern(TypePattern::FFIErrorEnum(e)) => {
                 indented!(w, [_], r#"var rval = {};"#, fn_call)?;
@@ -971,61 +913,6 @@ pub trait CSharpWriter {
         Ok(())
     }
 
-    fn write_pattern_service_success_enum_aware_rval(&self, w: &mut IndentWriter, _class: &Service, function: &Function, deref_context: bool) -> Result<(), Error> {
-        self.debug(w, "write_pattern_service_success_enum_aware_rval")?;
-        let mut args = self.pattern_class_args_without_first_to_string(function, false);
-
-        // Make sure we don't have a `,` when only single parameter
-        if !args.is_empty() {
-            args = format!(", {}", args);
-        }
-
-        let context = if deref_context { "_context".to_string() } else { "ref _context".to_string() };
-
-        match function.signature().rval() {
-            CType::Pattern(TypePattern::FFIErrorEnum(e)) => {
-                indented!(w, r#"var rval = {}.{}({} {});"#, self.config().class, function.name(), context, args)?;
-                indented!(w, r#"if (rval != {}.{})"#, e.the_enum().rust_name(), e.success_variant().name())?;
-                indented!(w, r#"{{"#)?;
-                indented!(w, [_], r#"throw new Exception($"Something went wrong {{rval}}");"#)?;
-                indented!(w, r#"}}"#)?;
-            }
-            CType::Primitive(PrimitiveType::Void) => {
-                indented!(w, r#"{}.{}({}{});"#, self.config().class, function.name(), context, args)?;
-            }
-            _ => {
-                indented!(w, r#"return {}.{}({}{});"#, self.config().class, function.name(), context, args)?;
-            }
-        }
-
-        Ok(())
-    }
-
-    fn pattern_class_args_without_first_to_string(&self, function: &Function, with_types: bool) -> String {
-        function
-            .signature()
-            .params()
-            .iter()
-            .skip(1)
-            .map(|x| {
-                let with_type = self.converter().to_typespecifier_in_param(x.the_type());
-
-                let name = if with_types {
-                    x.name().to_string()
-                } else if with_type.contains("ref ") {
-                    format!("ref {}", x.name().to_string())
-                } else if with_type.contains("out ") {
-                    format!("out {}", x.name().to_string())
-                } else {
-                    x.name().to_string()
-                };
-
-                format!("{}{}", if with_types { format!("{} ", with_type) } else { "".to_string() }, name)
-            })
-            .collect::<Vec<_>>()
-            .join(", ")
-    }
-
     fn write_all(&self, w: &mut IndentWriter) -> Result<(), Error> {
         self.write_file_header_comments(w)?;
         w.newline()?;
